# coding=utf-8
# Copyright 2020 The HuggingFace Inc. team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from typing import List, Optional, Union, TYPE_CHECKING

import torch
import torch.distributed as dist
from torch import nn

from transformers.generation.configuration_utils import GenerationConfig

from .sbs_helpers.gumbel import gumbel, gumbel_with_maximum
from .sbs_helpers.logits_process import LogitsProcessorList, LogitsProcessor
from transformers.generation.stopping_criteria import StoppingCriteriaList, validate_stopping_criteria
from .beam_constrained_decoder import ConstrainedDecoder
from .contrastive_decoder import ContrastiveDecoder
from .utils import GenerationStrategy, BeamSearchDecoderOnlyOutput, BeamSearchEncoderDecoderOutput, \
    BeamSearchOutput
from .beam_utils import BeamSearchScorer

if TYPE_CHECKING:
    from transformers.modeling_utils import PreTrainedModel, GenerationMixin
    from transformers.generation.streamers import BaseStreamer


class OldSBSLogitProcessor(LogitsProcessor):

    def __init__(self, num_beams, batch_size=2):
        self.num_beams = num_beams
        self.batch_size = batch_size

    def __call__(self, input_ids, scores, **kwargs):
        """

        :param input_ids: shape (batch_size * num_beams, seq_len)
        :param scores: shape (batch_size * num_beams, vocab_size)
        :param kwargs.beam_log_probs: shape (batch_size * num_beams,)
        :param kwargs.past_processed_scores: tuple of size seq_len with tensors (batch_size * num_beams, vocab_size)
        :return: shape (batch_size, vocab_size)
        """

        assert kwargs.get('beam_log_probs', None) is not None, "beam_log_probs must be provided"
        assert kwargs.get('past_scores', None) is not None, "past_scores is None. Add output_" \
                                                                      "scores=True to the generate function arguments"
        assert kwargs.get('beam_indices', None) is not None, "beam_indices must be provided"
        beam_log_probs = kwargs['beam_log_probs']  # shape (batch_size * num_beams,)
        beam_log_probs = beam_log_probs.view(-1,1)  # shape (batch_size * num_beams,)
        past_scores = kwargs['past_scores']  # tuple: seq_len tensors of dim (batch_size * num_beams, vocab_size)
        beam_indices = kwargs['beam_indices']  # shape (batch_size * num_beams,)

        device = input_ids.device

        scores = beam_log_probs.view(-1, 1) + scores  # shape (batch_size * num_beams, vocab_size)
        # scores = scores.clamp(min=-1e9)

        if len(past_scores) == 0:  # first token
            last_gumbels = gumbel(size=(self.batch_size,)).to(device)  # shape (batch_size, )
            # expand (batch_size, ) to (batch_size * num_beams, )
            last_gumbels = last_gumbels.repeat_interleave(self.num_beams, dim=0)
        else:
            last_token_scores = past_scores[-1]  # shape (batch_size * num_beams, vocab_size)
            last_tokens = input_ids[:, -1]  # shape (batch_size * num_beams, )
            last_beam_indices = torch.tensor(tuple(tup[-1] for tup in beam_indices), device=device)
            last_gumbels = last_token_scores[last_beam_indices, last_tokens]
        new_gumbels, _ = gumbel_with_maximum(scores, last_gumbels)  # shape (batch_size * num_beams, ) todo check this view
        return new_gumbels


class OldStochasticBeamSearchDecoder(GenerationStrategy):
    """
    Implements beam search decoding method.
    """

    def __init__(self, config: GenerationConfig = None, check_config: bool = True):
        super().__init__(config, check_config=check_config)
        self.scorer = None

    @classmethod
    def param_check(cls, config: GenerationConfig):
        if config.num_beams <= 1:
            raise ValueError(
                f"{cls.__name__} cannot be used with num_beams <= 1, but num_beams is {config.num_beams}."
            )
        if config.num_beam_groups != 1:
            raise ValueError(
                f"{cls.__name__} cannot be used with num_beam_groups != 1, "
                f"but num_beam_groups is {config.num_beam_groups}."
            )
        if config.do_sample:
            raise ValueError(
                f"{cls.__name__} cannot be used with sampling."
            )
        if ConstrainedDecoder.param_test(config):
            raise ValueError(
                f"{cls.__name__} cannot be used with constraints."
            )
        if ContrastiveDecoder.param_test(config):
            raise ValueError(
                f"{cls.__name__} cannot be used with contrastive penalties."
            )

    def __call__(self,
                 model: Union["PreTrainedModel", "GenerationMixin"],
                 input_ids: torch.LongTensor,
                 assistant_model: Optional["PreTrainedModel"] = None,
                 top_k: Optional[int] = None,
                 top_p: Optional[float] = None,
                 typical_p: Optional[float] = None,
                 penalty_alpha: Optional[float] = None,
                 do_sample: bool = False,
                 logits_processor: Optional[LogitsProcessorList] = None,
                 logits_warper: Optional[LogitsProcessorList] = None,
                 max_length: Optional[int] = None,
                 stopping_criteria: Optional[StoppingCriteriaList] = None,
                 pad_token_id: Optional[int] = None,
                 eos_token_id: Optional[Union[int, List[int]]] = None,
                 output_attentions: Optional[bool] = None,
                 output_hidden_states: Optional[bool] = None,
                 output_scores: Optional[bool] = True,
                 return_dict_in_generate: Optional[bool] = True,
                 synced_gpus: bool = False,
                 streamer: Optional["BaseStreamer"] = None,
                 sequential: Optional[bool] = None,
                 beam_scores_type: Optional[torch.dtype] = torch.float,
                 **model_kwargs,
                 ) -> Union[BeamSearchOutput, torch.LongTensor]:
        r"""
        Generates sequences of token ids for models with a language modeling head using **beam search decoding** and
        can be used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.

        <Tip warning={true}>

        In most cases, you do not need to call [`~generation.GenerationMixin.beam_search`] directly. Use generate()
        instead. For an overview of generation strategies and code examples, check the [following
        guide](../generation_strategies).

        </Tip>

        Parameters:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                The sequence used as a prompt for the generation.
            beam_scorer (`transformers.generation.BeamScorer`):
                An derived instance of [`BeamScorer`] that defines how beam hypotheses are constructed, stored and
                sorted during generation. For more information, the documentation of [`BeamScorer`] should be read.
            logits_processor (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
                used to modify the prediction scores of the language modeling head applied at each generation step.
            stopping_criteria (`StoppingCriteriaList`, *optional*):
                An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
                used to tell if the generation loop should stop.
            max_length (`int`, *optional*, defaults to 20):
                **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
                tokens. The maximum length of the sequence to be generated.
            pad_token_id (`int`, *optional*):
                The id of the *padding* token.
            eos_token_id (`Union[int, List[int]]`, *optional*):
                The id of the *end-of-sequence* token. Optionally, use a list to set multiple *end-of-sequence* tokens.
            output_attentions (`bool`, *optional*, defaults to `False`):
                Whether or not to return the attentions tensors of all attention layers. See `attentions` under
                returned tensors for more details.
            output_hidden_states (`bool`, *optional*, defaults to `False`):
                Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
                for more details.
            output_scores (`bool`, *optional*, defaults to `False`):
                Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
            return_dict_in_generate (`bool`, *optional*, defaults to `False`):
                Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
            synced_gpus (`bool`, *optional*, defaults to `False`):
                Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
            model_kwargs:
                Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
                an encoder-decoder model the kwargs should include `encoder_outputs`.

        Return:
            [`generation.BeamSearchDecoderOnlyOutput`], [`~generation.BeamSearchEncoderDecoderOutput`] or
            `torch.LongTensor`: A `torch.LongTensor` containing the generated tokens (default behaviour) or a
            [`~generation.BeamSearchDecoderOnlyOutput`] if `model.config.is_encoder_decoder=False` and
            `return_dict_in_generate=True` or a [`~generation.BeamSearchEncoderDecoderOutput`] if
            `model.config.is_encoder_decoder=True`.


        Examples:

        ```python
        >>> from transformers.generation import BeamSearchScorer        >>> from transformers import (
        ...     AutoTokenizer,
        ...     AutoModelForSeq2SeqLM,
        ...     LogitsProcessorList,
        ...     MinLengthLogitsProcessor,
        ...     )
        >>> import torch

        >>> tokenizer = AutoTokenizer.from_pretrained("t5-base")
        >>> model = AutoModelForSeq2SeqLM.from_pretrained("t5-base")

        >>> encoder_input_str = "translate English to German: How old are you?"
        >>> encoder_input_ids = tokenizer(encoder_input_str, return_tensors="pt").input_ids


        >>> # lets run beam search using 3 beams
        >>> num_beams = 3
        >>> # define decoder start token ids
        >>> input_ids = torch.ones((num_beams, 1), device=model.device, dtype=torch.long)
        >>> input_ids = input_ids * model.config.decoder_start_token_id

        >>> # add encoder_outputs to model keyword arguments
        >>> model_kwargs = {
        ...     "encoder_outputs": model.get_encoder()(
        ...         encoder_input_ids.repeat_interleave(num_beams, dim=0), return_dict=True
        ...     )
        ... }

        >>> # instantiate beam scorer
        >>> beam_scorer = BeamSearchScorer(
        ...     batch_size=1,
        ...     num_beams=num_beams,
        ...     device=model.device,
        ... )

        >>> # instantiate logits processors
        >>> logits_processor = LogitsProcessorList(
        ...     [
        ...         MinLengthLogitsProcessor(5, eos_token_id=model.config.eos_token_id),
        ...     ]
        ... )

        >>> outputs = model.beam_search(input_ids, beam_scorer, logits_processor=logits_processor, **model_kwargs)

        >>> tokenizer.batch_decode(outputs, skip_special_tokens=True)
        ['Wie alt bist du?']
        ```"""
        if self.config.num_beams <= 1:
            raise ValueError(
                f"{self.__class__.__name__} cannot be used with `num_beams == 1`. Please use `num_beams > 1`."
            )
        if logits_processor is not None:
            logits_processor = LogitsProcessorList(logits_processor)
            if any(isinstance(p, OldSBSLogitProcessor) for p in logits_processor):
                pass
            else:
                logits_processor.append(OldSBSLogitProcessor(num_beams=self.config.num_beams, batch_size=input_ids.shape[0]))
        else:
            logits_processor = LogitsProcessorList([OldSBSLogitProcessor(num_beams=self.config.num_beams,
                                                                         batch_size=input_ids.shape[0])])

        if self.check_config:
            if self.config.num_return_sequences > self.config.num_beams:
                raise ValueError("`num_return_sequences` has to be smaller or equal to `num_beams`.")

            if stopping_criteria.max_length is None:
                raise ValueError("`max_length` needs to be a stopping_criteria for now.")

        batch_size = input_ids.shape[0]
        # 11. prepare beam search scorer
        if self.scorer is None:
            self.scorer = BeamSearchScorer(
                batch_size=batch_size,
                num_beams=self.config.num_beams,
                device=input_ids.device,
                length_penalty=self.config.length_penalty,
                do_early_stopping=self.config.early_stopping,
                num_beam_hyps_to_keep=self.config.num_return_sequences,
                max_length=self.config.max_length,
            )
        # 12. interleave input_ids with `num_beams` additional sequences per batch
        input_ids, model_kwargs = model._expand_inputs_for_generation(
            input_ids=input_ids,
            expand_size=self.config.num_beams,
            is_encoder_decoder=model.config.is_encoder_decoder,
            **model_kwargs,
        )
        # 13. run beam search

        # init values
        logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
        stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
        if max_length is not None:
            warnings.warn(
                "`max_length` is deprecated in this function, use"
                " `stopping_criteria=StoppingCriteriaList(MaxLengthCriteria(max_length=max_length))` instead.",
                UserWarning,
            )
            stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
        if len(stopping_criteria) == 0:
            warnings.warn("You don't have defined any stopping_criteria, this will likely loop forever", UserWarning)
        pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
        eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
        if isinstance(eos_token_id, int):
            eos_token_id = [eos_token_id]
        output_scores = output_scores if output_scores is not None else self.config.output_scores
        output_attentions = (
            output_attentions if output_attentions is not None else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict_in_generate = (
            return_dict_in_generate
            if return_dict_in_generate is not None
            else self.config.return_dict_in_generate
        )

        batch_size = len(self.scorer._beam_hyps)
        num_beams = self.scorer.num_beams

        batch_beam_size, cur_len = input_ids.shape

        if num_beams * batch_size != batch_beam_size:
            raise ValueError(
                f"Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."
            )

        # init attention / hidden states / scores tuples
        scores = () if (return_dict_in_generate and output_scores) else None
        beam_indices = (
            tuple(() for _ in range(batch_beam_size)) if (return_dict_in_generate and output_scores) else None
        )
        decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
        cross_attentions = () if (return_dict_in_generate and output_attentions) else None
        decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

        # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
        if return_dict_in_generate and model.config.is_encoder_decoder:
            encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
            encoder_hidden_states = (
                model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
            )

        # initialise score of first beam with 0 and the rest with -1e9. This makes sure that only tokens
        # of the first beam are considered to avoid sampling the exact same tokens across all beams.
        beam_scores = torch.zeros((batch_size, num_beams), dtype=beam_scores_type, device=input_ids.device)
        beam_scores[:, 1:] = -float("inf")
        beam_scores = beam_scores.view((batch_size * num_beams,))

        this_peer_finished = False  # used by synced_gpus only
        while True:
            if synced_gpus:
                # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
                # The following logic allows an early break if all peers finished generating their sequence
                this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(input_ids.device)
                # send 0.0 if we finished, 1.0 otherwise
                dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
                # did all peers finish? the reduced sum will be 0.0 then
                if this_peer_finished_flag.item() == 0.0:
                    break

            model_inputs = model.prepare_inputs_for_generation(input_ids, **model_kwargs)

            outputs = model(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
            )

            if synced_gpus and this_peer_finished:
                cur_len = cur_len + 1
                continue  # don't waste resources running the code we don't need

            next_token_logits = outputs.logits[:, -1, :]
            next_token_scores = nn.functional.log_softmax(
                next_token_logits, dim=-1
            )  # (batch_size * num_beams, vocab_size)

            model_log_probs = next_token_scores.clone()
            next_token_scores_processed = logits_processor(input_ids,
                                                           next_token_scores,
                                                           beam_log_probs=beam_scores,
                                                           past_scores=scores,
                                                           beam_indices=beam_indices, )
            next_token_scores = next_token_scores_processed
            # Store scores, attentions and hidden_states when required
            if return_dict_in_generate:
                if output_scores:
                    scores += (next_token_scores_processed,)
                if output_attentions:
                    decoder_attentions += (
                        (outputs.decoder_attentions,) if model.config.is_encoder_decoder else (outputs.attentions,)
                    )
                    if model.config.is_encoder_decoder:
                        cross_attentions += (outputs.cross_attentions,)

                if output_hidden_states:
                    decoder_hidden_states += (
                        (outputs.decoder_hidden_states,)
                        if model.config.is_encoder_decoder
                        else (outputs.hidden_states,)
                    )

            # reshape for beam search
            vocab_size = next_token_scores.shape[-1]
            next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)

            # Sample 2 next tokens for each beam (so we have some spare tokens and match output of beam search)
            next_token_scores, next_tokens = torch.topk(
                next_token_scores, 2 * num_beams, dim=1, largest=True, sorted=True
            )

            model_log_probs = model_log_probs.view((batch_size, num_beams * vocab_size)).gather(dim=1,
                                                                                                index=next_tokens)

            next_indices = torch.div(next_tokens, vocab_size, rounding_mode="floor")
            next_tokens = next_tokens % vocab_size

            # stateless
            beam_outputs = self.scorer.process(
                input_ids,
                next_token_scores,
                next_tokens,
                next_indices,
                pad_token_id=pad_token_id,
                eos_token_id=eos_token_id,
                beam_indices=beam_indices,
                beam_scores=beam_scores,
                next_true_log_probs=model_log_probs,
            )

            beam_scores = beam_outputs["next_beam_scores"]
            beam_next_tokens = beam_outputs["next_beam_tokens"]
            beam_idx = beam_outputs["next_beam_indices"]
            beam_gumbels = beam_outputs["next_beam_gumbels"]

            input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)

            model_kwargs = model._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=model.config.is_encoder_decoder
            )
            if model_kwargs["past_key_values"] is not None:
                model_kwargs["past_key_values"] = model._reorder_cache(model_kwargs["past_key_values"], beam_idx)

            if return_dict_in_generate and output_scores:
                beam_indices = tuple((beam_indices[beam_idx[i]] + (beam_idx[i],) for i in range(len(beam_indices))))

            # increase cur_len
            cur_len = cur_len + 1

            if self.scorer.is_done or stopping_criteria(input_ids, scores):
                if not synced_gpus:
                    break
                else:
                    this_peer_finished = True

        sequence_outputs = self.scorer.finalize(
            input_ids,
            beam_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            max_length=stopping_criteria.max_length,
            beam_indices=beam_indices,
            beam_gumbels=beam_gumbels,
        )

        if return_dict_in_generate:
            if not output_scores:
                sequence_outputs["sequence_scores"] = None

            if model.config.is_encoder_decoder:
                return BeamSearchEncoderDecoderOutput(
                    sequences=sequence_outputs["sequences"],
                    sequences_scores=sequence_outputs["sequence_scores"],
                    scores=scores,
                    beam_indices=sequence_outputs["beam_indices"],
                    encoder_attentions=encoder_attentions,
                    encoder_hidden_states=encoder_hidden_states,
                    decoder_attentions=decoder_attentions,
                    cross_attentions=cross_attentions,
                    decoder_hidden_states=decoder_hidden_states,
                    beam_gumbels=sequence_outputs["beam_gumbels"],
                )
            else:
                return BeamSearchDecoderOnlyOutput(
                    sequences=sequence_outputs["sequences"],
                    sequences_scores=sequence_outputs["sequence_scores"],
                    scores=scores,
                    beam_indices=sequence_outputs["beam_indices"],
                    attentions=decoder_attentions,
                    hidden_states=decoder_hidden_states,
                    beam_gumbels=sequence_outputs["beam_gumbels"],
                )
        else:
            return sequence_outputs["sequences"]



from typing import List, Union, Optional

import torch
from transformers import StoppingCriteriaList

from .beam_search import BeamSearchDecoder
from .simple_beam_utils import BeamSearchNode
from ..strategies.sbs_helpers.gumbel import gumbel, gumbel_with_maximum
from ..strategies.sbs_helpers.logits_process import LogitsProcessor, LogitsProcessorList


class SimpleSBSLogitProcessor(LogitsProcessor):

    def __call__(self, input_ids, scores, **kwargs):
        """
        LogitsProcessor that implements stochastic beam search.

        :param input_ids: shape (batch_size * num_beams, seq_len)
        :param scores: shape (batch_size * num_beams, vocab_size)
        :param kwargs.beam_log_probs: shape (batch_size * num_beams,)
        :param kwargs.past_processed_scores: tuple of size seq_len with tensors (batch_size * num_beams, vocab_size)
        :return: shape (batch_size, vocab_size)
        """
        batch_size, seq_len = input_ids.shape
        device = input_ids.device

        if kwargs["nodes"] is None:
            # no nodes -> we are in the first step of generation. input_ids is of shape (batch_size, 1)
            last_gumbels = gumbel(size=(batch_size,)).to(input_ids.device)  # shape (batch_size, )
            log_probs = scores
        else:
            # input_ids is of shape (alive_beams,seq_len)
            nodes: List[BeamSearchNode] = kwargs['nodes']  # list of size num_active_beams
            last_gumbels = torch.tensor([node.last_score for node in nodes], device=device)
            beam_log_probs = torch.tensor([node.log_prob for node in nodes], device=device)
            log_probs = scores + beam_log_probs.unsqueeze(-1)  # shape (num_active_beams, vocab_size)

        new_gumbels, _ = gumbel_with_maximum(log_probs, last_gumbels)  # shape (num_active_beams, vocab_size)
        return new_gumbels


class StochasticBeamSearchDecoder(BeamSearchDecoder):
    
    def __repr__(self):
        return f"simple.{self.__class__.__name__}()"

    def __call__(self,
                 model: Union["PreTrainedModel", "GenerationMixin"],
                 input_ids: torch.LongTensor,
                 logits_processor: Optional[LogitsProcessorList] = None,
                 stopping_criteria: Optional[StoppingCriteriaList] = None,
                 keep_k_always_alive: Optional[int] = False,
                 disable_kv_cache: Optional[bool] = False,
                 eval_by_score: Optional[bool] = True,
                 encoder_input_ids: Optional[torch.LongTensor] = None,
                 pad_token_id: Optional[int] = None,
                 **model_kwargs,
                 ):
        r"""
        Simple wrapper around the Beam Search decoder that appends the stochastic beam search logits processor.
        """
        if logits_processor is not None:
            logits_processor = LogitsProcessorList(logits_processor)
            if any(isinstance(p, SimpleSBSLogitProcessor) for p in logits_processor):
                pass
            else:
                logits_processor.append(SimpleSBSLogitProcessor())
        else:
            logits_processor = LogitsProcessorList([SimpleSBSLogitProcessor()])
        return super().__call__(model, input_ids, logits_processor, stopping_criteria, keep_k_always_alive,
                                disable_kv_cache, eval_by_score, encoder_input_ids, pad_token_id,
                                **model_kwargs)



import torch

def _prepare_decoder_input(model, sequences):
    batch_size = sequences.shape[0]
    device = sequences.device
    start_token = model.config.decoder_start_token_id
    assert torch.all(sequences[:, 0] == start_token)
    decoder_input = torch.tensor([[start_token]], dtype=torch.long, device=device).repeat(batch_size, 1)
    return decoder_input.view(batch_size, 1)


def compute_transition_scores_nonbeam(sequences, scores, normalize_logits=True):
    if len(sequences.shape) == 1:
        sequences = sequences.unsqueeze(0)
    scores = torch.stack(scores, dim=1) # batch_size x seq_len x vocab_size
    if normalize_logits:
        scores = scores.log_softmax(dim=-1)
    sequences = sequences[:, 1:]
    transition_scores = scores.gather(dim=-1, index=sequences.unsqueeze(-1)).squeeze(-1)
    return transition_scores


def compute_true_logprobs(model, sequences, encoder_input=None):
    if encoder_input is None:
        model_kwargs = {}
    else:
        encoder_input, model_input_name, model_kwargs = model._prepare_model_inputs(
            encoder_input.input_ids, model.config.bos_token_id, {'attention_mask': encoder_input.attention_mask},
        )
        model_kwargs = model._prepare_encoder_decoder_kwargs_for_generation(
            encoder_input, model_kwargs, model_input_name
        )
        _, model_kwargs = model._expand_inputs_for_generation(sequences.shape[0] // encoder_input.shape[0],
                                                              is_encoder_decoder=True, **model_kwargs)

    device = sequences.device
    if len(sequences.shape) == 1:
        sequences = sequences.unsqueeze(0)
    batch_size, seq_len = sequences.shape
    logp = torch.zeros(batch_size, device=device)
    transition_scores = torch.zeros(batch_size, seq_len - 1, device=device)
    finished = torch.zeros(batch_size, device=device, dtype=torch.bool)
    with (torch.no_grad()):
        decoder_input_ids = _prepare_decoder_input(model, sequences)
        # decoder_input_ids = torch.tensor([[model.config.decoder_start_token_id]], dtype=torch.long, device=device)
        # assert torch.all(decoder_input_ids == alt_decoder_input_ids)
        for i in range(seq_len - 1):
            outputs = model(**model_kwargs, decoder_input_ids=decoder_input_ids)
            last_logits = outputs.logits[:, -1, :]
            last_logits = last_logits.log_softmax(dim=-1)
            next_token = sequences[:, i+1].view(batch_size,)
            logp += last_logits.gather(dim=-1, index=next_token.unsqueeze(-1)).squeeze(-1) * (~finished).float()
            transition_scores[:, i] = last_logits.gather(dim=-1, index=next_token.unsqueeze(-1)).squeeze(-1) * (~finished).float()
            finished = finished | (next_token == model.config.eos_token_id)
            decoder_input_ids = torch.cat([decoder_input_ids, next_token.unsqueeze(-1)], dim=1)
    return logp, transition_scores


def compute_logprobs_from_scores(model, outputs):
    logp_computed = model.compute_transition_scores(sequences=outputs.sequences, scores=outputs.scores, normalize_logits=True)
    nonpad_mask = outputs.sequences[:, 1:] != model.config.pad_token_id
    result = torch.where(nonpad_mask, logp_computed, torch.zeros_like(logp_computed))
    return result


def test_t5():
    from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    inputs = tokenizer(["translate English to German: What is your name, my dear Friend? I missed you so much",
                           "translate English to German: How old are you?",
                          ],
                          return_tensors="pt", padding=True, truncation=True
                          )
    outputs = model.generate(**inputs, num_beams=1, do_sample=True, num_return_sequences=1,
                             return_dict_in_generate=True, output_scores=True,
                             top_k=0,
                             # length_penalty=0.0, early_stopping=True,
                             max_new_tokens=100)
    # gen_logp = outputs.sequences_scores
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences)}")
    print(f"generated tokens: {outputs.sequences}")
    # print(f"generated logp: {gen_logp}")
    _, gold_transitions = compute_true_logprobs(model, outputs.sequences, encoder_input=inputs)
    # logp_computed = model.compute_transition_scores(sequences=outputs.sequences, scores=outputs.scores,
    #                                                 beam_indices=outputs.beam_indices,
    #                                                 normalize_logits=True)
    logp_computed = compute_logprobs_from_scores(model, outputs)
    print(f"score transitions: {logp_computed.sum(dim=1)}")
    print(f"gold transitions:  {gold_transitions.sum(dim=1)}")



if __name__ == "__main__":
    test_t5()



from itertools import count
from queue import Queue
from heapq import heappush, heappop
from typing import Dict, Any, List

import torch
from torch.nn.functional import pad


class BSPriorityQueue(Queue):
    """ This PQ has the following customizations for beam search:
     - If there is a tie, FIFO using a counter
     - minheap -> maxheap (by negating scores)
     - gets score from node.eval()
     - defines merge operation
     - defines peek operation
     - defines get_n operation
     """

    def _init(self, maxsize):
        self.queue = []
        self.count = count()

    def _qsize(self):
        return len(self.queue)

    def _put(self, item):
        score = item.eval()
        heappush(self.queue, (-score, next(self.count), item))

    def _get(self):
        score, _, item = heappop(self.queue)
        return item

    def peek(self):
        return self.queue[0][2]

    def get_n(self, n):
        items = []
        for _ in range(n):
            if not self.empty():
                items.append(self.get())
            else:
                break
        return items

    @classmethod
    def merge(cls, q1, q2):
        new = BSPriorityQueue()
        for _ in range(q1.qsize()):
            node = q1.get()
            new.put(node)
        for _ in range(q2.qsize()):
            node = q2.get()
            new.put(node)
        return new

    def __repr__(self):
        return f"BSPriorityQueue(len={len(self.queue)}, data={str(self.queue)})"


class SimpleBeamSearch(object):
    """ Defines a beam search object for a single input sentence. """

    def __init__(self, beam_size, eos_token, eval_by_score=False):
        self.beam_size = beam_size
        self.eos_token = eos_token
        self.eval_by_score = eval_by_score

        self.alive_nodes = BSPriorityQueue()  # beams to be expanded
        self.final_nodes = BSPriorityQueue()  # beams that ended in EOS

    def add(self, node, finished=False):
        """ Adds a new beam search node to the queue of current nodes """
        if node.log_prob == float("-inf"):
            return  # this is the only hack in the algorithm. Discard nodes with -inf log prob. Necessary for toy models
        if finished:
            self.final_nodes.put(node)
        else:
            self.alive_nodes.put(node)

    def get_current_beams(self):
        """ Returns beam_size current nodes with the lowest negative log probability """
        return self.alive_nodes.get_n(self.beam_size)

    def final_best_n(self, number):
        """ Returns final nodes with the lowest negative log probability """
        # Merge EOS paths and those that were stopped by
        # max sequence length (still in nodes)
        merged = BSPriorityQueue.merge(self.final_nodes, self.alive_nodes)
        best = merged.get_n(number)  # todo: make more efficient by peeking both queues instead of merge
        return best

    def get_best_node(self):
        """ Returns and removes from queue the best node. True if final, False otherwise"""
        best_alive = self.alive_nodes.peek() if not self.alive_nodes.empty() else None
        best_final = self.final_nodes.peek() if not self.final_nodes.empty() else None
        best_alive_score = best_alive.eval() if best_alive is not None else float("-inf")
        best_final_score = best_final.eval() if best_final is not None else float("-inf")
        if best_alive is None and best_final is None:
            return None, None
        if (best_alive is None) or (best_alive_score < best_final_score):
            return self.final_nodes.get(), True
        else:
            return self.alive_nodes.get(), False

    def prune(self, keep_k_always_alive=False):
        """ Removes all nodes but the beam_size best ones (lowest neg log prob) """
        new_nodes_active = BSPriorityQueue()
        new_nodes_final = BSPriorityQueue()
        if keep_k_always_alive:  # this is a HACK to mimic HF's early stopping. Should be False for correctness.
            for _ in range(self.beam_size):
                if not self.alive_nodes.empty():
                    node = self.alive_nodes.get()
                    new_nodes_active.put(node)
                if not self.final_nodes.empty():
                    node = self.final_nodes.get()
                    new_nodes_final.put(node)
        else:
            for _ in range(self.beam_size): #TODO: this can be optimized by discarding worse-than-worst nodes on the go
                node, is_final = self.get_best_node()
                if node is None:
                    break
                if not is_final:
                    new_nodes_active.put(node)
                else:
                    new_nodes_final.put(node)
        self.alive_nodes = new_nodes_active
        self.final_nodes = new_nodes_final

    def is_done(self):
        """ Returns whether beam search is complete or not """
        # best_node_final = len(self.final_nodes.peek().sequence) if not self.final_nodes.empty() else 0
        # print((len(self.final_nodes.queue), len(self.alive_nodes.queue),best_node_final ))
        return len(self.final_nodes.queue) >= self.beam_size

    def __repr__(self):
        return f"BSSearch({len(self.alive_nodes.queue)} alive, {len(self.final_nodes.queue)} final)"


class BeamSearchNode(object):
    """ Defines a search node and stores values important for computation of beam search path"""

    def __init__(self, search, sequence, model_state, log_prob, last_score):
        self.search = search
        self.sequence = sequence
        self.model_state = model_state
        self.log_prob = log_prob
        self.last_score = last_score

    def eval(self):
        if self.search.eval_by_score:
            return self.last_score
        else:
            return self.log_prob

    def __repr__(self):
        return f"BeamSearchNode(logp={self.log_prob.item():.2f}, score={self.last_score.item():.2f}, seq={self.sequence})"


# In the following ugly code rests the decent speed of this beam search implementation.
def collate_model_states(model_states, disable_kv_cache=False):
    new_dict = model_states[0].copy()
    for key in model_states[0]:
        if model_states[0][key] is not None and isinstance(model_states[0][key], torch.Tensor):
            new_dict[key] = torch.stack([model_states[i][key].squeeze() for i in range(len(model_states))], dim=0)
    if model_states[0].get("encoder_outputs") is not None:
        new_dict["encoder_outputs"] = collate_model_states(
            [model_states[i]["encoder_outputs"] for i in range(len(model_states))])
    if model_states[0].get("past_key_values") is not None and not disable_kv_cache:
        tt = []
        for i in range(len(model_states[0]["past_key_values"])):
            list_t = []
            for j in range(len(model_states[0]["past_key_values"][i])):
                t = torch.stack([model_states[k]["past_key_values"][i][j] for k in range(len(model_states))], dim=0)
                list_t.append(t)
            tt.append(tuple(list_t))
        new_dict["past_key_values"] = tuple(tt)
    if disable_kv_cache:
        new_dict.pop("past_key_values", None)
    return new_dict


def update_model_kv_cache(model_state, index, model_kwargs, disable_kv_cache=False):
    new_dict = model_state.copy()
    if not disable_kv_cache:
        new_dict["past_key_values"] = tuple(
            [tuple([a[index] for a in b]) for b in model_kwargs["past_key_values"]])
    else:
        new_dict.pop("past_key_values", None)
    new_dict["attention_mask"] = model_kwargs["attention_mask"][index].unsqueeze(0)
    return new_dict


def separate_model_states(
        batch_size: int,
        is_encoder_decoder: bool = False,
        disable_kv_cache: bool = False,
        new_method=False,
        **model_kwargs,
) -> List[Dict[str, Any]]:
    """  Separate the batch of encoder states into a list of individual states. """

    def _separate_dict(dict_to_expand, i):
        new_dict = dict_to_expand.copy()
        for key in dict_to_expand:
            if dict_to_expand[key] is not None and isinstance(dict_to_expand[key], torch.Tensor):
                new_dict[key] = dict_to_expand[key][i].unsqueeze(0)
        return new_dict

    model_kwargs_l = [_separate_dict(model_kwargs, i) for i in range(batch_size)]

    if is_encoder_decoder:
        if model_kwargs.get("encoder_outputs") is None:
            raise ValueError("If `is_encoder_decoder` is True, make sure that `encoder_outputs` is defined.")
        for i in range(batch_size):
            model_kwargs_l[i]["encoder_outputs"] = _separate_dict(model_kwargs["encoder_outputs"], i)
        if model_kwargs.get("past_key_values") is not None and not disable_kv_cache:
            for i in range(batch_size):
                if new_method:
                    pass #recursive_replace_tensors(model_kwargs_l[i]["past_key_values"], i)
                else:
                    model_kwargs_l[i]["past_key_values"] = tuple(
                        [tuple([a[i] for a in b]) for b in model_kwargs["past_key_values"]])
    else:
        if model_kwargs.get("past_key_values") is not None and not disable_kv_cache:
            for i in range(batch_size):
                if new_method:
                    pass #recursive_replace_tensors(model_kwargs_l[i]["past_key_values"], i)
                else:
                    model_kwargs_l[i]["past_key_values"] = tuple(
                        [tuple([a[i] for a in b]) for b in model_kwargs["past_key_values"]])
    return model_kwargs_l


def pad_tensors(tensors, pad_token_id):
    if len(tensors[0].shape) > 0:
        max_len = max([t.shape[0] for t in tensors])
        tensors = [pad(t, (0, max_len - t.shape[0]), value=pad_token_id) for t in tensors]
        return tensors
    raise ValueError("Tensors must have at least two dimensions")



from typing import Optional, Union, TYPE_CHECKING
import torch
from transformers import GenerationConfig, StoppingCriteriaList
from .simple_beam_utils import SimpleBeamSearch, BeamSearchNode, pad_tensors, separate_model_states, \
    collate_model_states, update_model_kv_cache
from ..strategies.sbs_helpers.logits_process import LogitsProcessorList
from ..strategies.utils import GenerationStrategy, BeamSearchDecoderOnlyOutput

if TYPE_CHECKING:
    from transformers.modeling_utils import PreTrainedModel, GenerationMixin


class BeamSearchDecoder(GenerationStrategy):

    def __init__(self, config: GenerationConfig = None, check_config: bool = True):
        super().__init__(config, check_config=check_config)

    @classmethod
    def param_check(cls, config: GenerationConfig):
        pass

    def __call__(self,
                 model: Union["PreTrainedModel", "GenerationMixin"],
                 input_ids: torch.LongTensor,
                 logits_processor: Optional[LogitsProcessorList] = None,
                 stopping_criteria: Optional[StoppingCriteriaList] = None,
                 keep_k_always_alive: Optional[int] = False,
                 disable_kv_cache: Optional[bool] = False,
                 eval_by_score: Optional[bool] = False,
                 encoder_input_ids: Optional[torch.LongTensor] = None,
                 pad_token_id: Optional[int] = None,
                 **model_kwargs,
                 ):
        r"""
        Simple implementation of Beam Search.

        Parameters:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                The sequence used as a prompt for the generation.
            logits_processor (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
                used to modify the prediction scores of the language modeling head applied at each generation step.
            stopping_criteria (`StoppingCriteriaList`, *optional*):
                An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
                used to tell if the generation loop should stop.
            keep_k_always_alive (`int`, *optional*, defaults to `False`):
                If set to `True`, always keep at least `num_beams` hypotheses alive, irrespective of number
                of finished hypotheses.
            model_kwargs:
                Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
                an encoder-decoder model the kwargs should include `encoder_outputs`.

        Return:
            [`~generation.SampleDecoderOnlyOutput`] or [`~generation.SampleEncoderDecoderOutput`].
        """
        from .stochastic_beam_search import SimpleSBSLogitProcessor
        if logits_processor is not None and any(isinstance(p, SimpleSBSLogitProcessor) for p in logits_processor) and not eval_by_score:
            print("WARNING: SimpleSBSLogitProcessor is used but eval_by_score is False.")

        logits_processor = LogitsProcessorList(logits_processor) if logits_processor is not None else LogitsProcessorList()
        stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()

        batch_size, _ = input_ids.shape
        num_beams = self.config.num_beams
        vocab_size = model.config.vocab_size if "vocab_size" in model.config.__dict__ else model.config.decoder.vocab_size
        pad_token_id = pad_token_id if pad_token_id is not None else pad_token_id
        
        # 0. Initialize beam searches, one for each sequence in the batch
        searches = [SimpleBeamSearch(num_beams, model.config.eos_token_id, eval_by_score) for _ in range(batch_size)]

        # 1. Generate initial hypotheses
        model_inputs = model.prepare_inputs_for_generation(input_ids, **model_kwargs)
        if encoder_input_ids is not None:
            model_inputs["input_ids"] = encoder_input_ids
        outputs = model(**model_inputs, return_dict=True)
        model_kwargs = model._update_model_kwargs_for_generation(outputs, model_kwargs,
                                                                 is_encoder_decoder=model.config.is_encoder_decoder)
        model_states = separate_model_states(batch_size, model.config.is_encoder_decoder, disable_kv_cache, **model_kwargs)

        logits = outputs.logits[:, -1, :]
        
        if self.__class__.__name__ == "StochasticBeamSearchDecoder":
            if len(logits_processor) > 1:
                prev_logit_processors = logits_processor[:-1]
                sbs_processor = logits_processor[-1]                    
                logits = LogitsProcessorList(prev_logit_processors)(input_ids, logits, nodes=None).log_softmax(dim=-1)
                scores = sbs_processor(input_ids, logits, nodes=None)
            else:
                logits = logits.log_softmax(dim=-1)
                scores = logits_processor(input_ids, logits, nodes=None)
        else:
            logits = logits.log_softmax(dim=-1)
            scores = logits_processor(input_ids, logits, nodes=None)
                    
        scores, next_candidates = torch.topk(scores, min(num_beams, vocab_size))  # (batch_size, num_beams)
        log_probs = torch.gather(logits, -1, next_candidates)  # order by scores, store true log probs

        # 1.1. Add initial hypotheses to the beam searches
        for b_idx in range(batch_size):
            for j in range(min(num_beams, vocab_size)):
                next_token = next_candidates[b_idx, j]
                new_seq = torch.cat((input_ids[b_idx], next_token.unsqueeze(0)))
                node = BeamSearchNode(searches[b_idx], new_seq, model_states[b_idx], log_probs[b_idx, j],
                                      scores[b_idx, j])
                searches[b_idx].add(node, finished=(next_token == model.config.eos_token_id))

        # 2. Beam search generation loop
        while True:
            # 2.1 Pop the current nodes to expand
            nodes = [n for s in searches for n in s.get_current_beams()]
            if len(nodes) == 0:
                break  # All beams ended in EOS

            # 2.2 Expand nodes, get top `num_beams` candidates
            input_ids = pad_tensors([node.sequence for node in nodes], pad_token_id)
            input_ids = torch.stack(input_ids, dim=0)
            model_args = collate_model_states([node.model_state for node in nodes], disable_kv_cache=disable_kv_cache)
            model_inputs = model.prepare_inputs_for_generation(input_ids, **model_args)
            if encoder_input_ids is not None:
                model_inputs["input_ids"] = encoder_input_ids
            outputs = model(**model_inputs, return_dict=True)
            
            model_args = model._update_model_kwargs_for_generation(outputs, model_args,
                                                                 is_encoder_decoder=model.config.is_encoder_decoder)
            #the call to update_model_kv_cache must update attention masks as in the prvious line

            logits = outputs.logits[:, -1, :]
            if self.__class__.__name__ == "StochasticBeamSearchDecoder":
                if len(logits_processor) > 1:
                    prev_logit_processors = logits_processor[:-1]
                    sbs_processor = logits_processor[-1]                    
                    logits = LogitsProcessorList(prev_logit_processors)(input_ids, logits, nodes=nodes).log_softmax(dim=-1)
                    scores = sbs_processor(input_ids, logits, nodes=nodes)
                else:
                    logits = logits.log_softmax(dim=-1)
                    scores = logits_processor(input_ids, logits, nodes=nodes)
            else:
                logits = logits.log_softmax(dim=-1)
                scores = logits_processor(input_ids, logits, nodes=nodes)
            scores, next_candidates = torch.topk(scores, min(num_beams, vocab_size))  # (batch_size, num_beams)
            log_probs = torch.gather(logits, -1, next_candidates)  # order by scores, store true log probs

            # 2.3 Push new candidates to the beam searches
            for n_idx, node in enumerate(nodes):
                model_state = update_model_kv_cache(node.model_state, n_idx, model_args, disable_kv_cache)
                for j in range(min(num_beams, vocab_size)):
                    next_token = next_candidates[n_idx, j]
                    beam_log_p = node.log_prob + log_probs[n_idx, j]
                    new_seq = torch.cat((input_ids[n_idx], next_token.unsqueeze(0)))
                    new_node = BeamSearchNode(node.search, new_seq, model_state, beam_log_p, scores[n_idx, j])
                    new_node.search.add(new_node, finished=(next_token == model.config.eos_token_id))

            # 2.4 Prune and check stopping criteria
            for search in searches:
                search.prune(keep_k_always_alive=keep_k_always_alive)

            if stopping_criteria(input_ids, None):
                break

            # rationale: if we always keep k non-eos beams alive, generation only stops by max_length
            # instead, with this heuristic, we mimic HF's early stopping: when we have k finished hypotheses, we allow
            # one more round of expansion to allow for EOS to be generated in alive beams.
            # THIS A HEURISTIC to test against HF's implementation. keep_k_always_alive should be False for correctness.
            if keep_k_always_alive and all([search.is_done() for search in searches]):
                keep_k_always_alive = False
        # 3. Return best hypotheses
        best_sents = []
        best_log_probs = []
        last_scores = []
        for search in searches:
            for node in search.final_best_n(self.config.num_return_sequences):
                best_sents.append(node.sequence.cpu())
                best_log_probs.append(node.log_prob.cpu().item())
                last_scores.append(node.last_score.cpu().item())

        best_sents = torch.stack(pad_tensors(best_sents, pad_token_id), dim=0)
        best_log_probs = torch.tensor(best_log_probs)
        last_scores = torch.tensor(last_scores)
        return BeamSearchDecoderOnlyOutput(
            sequences=best_sents,
            last_scores=last_scores,
            sequences_scores=best_log_probs,
        )



from typing import Optional, Union, TYPE_CHECKING
import torch
from transformers import GenerationConfig, StoppingCriteriaList, LogitsProcessorList

from ..strategies.utils import GenerationStrategy, SampleEncoderDecoderOutput

if TYPE_CHECKING:
    from transformers.modeling_utils import PreTrainedModel, GenerationMixin


class SamplingDecoder(GenerationStrategy):

    def __init__(self, config: GenerationConfig = None, check_config: bool = True):
        super().__init__(config, check_config=check_config)

    @classmethod
    def param_check(cls, config: GenerationConfig):
        pass

    def __call__(self,
                 model: Union["PreTrainedModel", "GenerationMixin"],
                 input_ids: torch.LongTensor,
                 logits_processor: Optional[LogitsProcessorList] = None,
                 logits_warper: Optional[LogitsProcessorList] = None,
                 stopping_criteria: Optional[StoppingCriteriaList] = None,
                 pad_token_id: Optional[int] = None,
                 quiet: Optional[bool] = True,
                 **model_kwargs,
                 ):
        r"""
        Generates ancestral sequences. May use logit sequences of token ids for models with a language modeling head using **multinomial sampling** and
        can be used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.

        Parameters:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                The sequence used as a prompt for the generation.
            logits_processor (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
                used to modify the prediction scores of the language modeling head applied at each generation step.
            stopping_criteria (`StoppingCriteriaList`, *optional*):
                An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
                used to tell if the generation loop should stop.
            logits_warper (`LogitsProcessorList`, *optional*):
                An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsWarper`] used
                to warp the prediction score distribution of the language modeling head applied before multinomial
                sampling at each generation step.
            model_kwargs:
                Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
                an encoder-decoder model the kwargs should include `encoder_outputs`.

        Return:
            [`~generation.SampleDecoderOnlyOutput`] or [`~generation.SampleEncoderDecoderOutput`].
"""

        # 12. expand input_ids with `num_return_sequences` additional sequences per batch
        input_ids, model_kwargs = model._expand_inputs_for_generation(
            input_ids=input_ids,
            expand_size=self.config.num_return_sequences,
            is_encoder_decoder=model.config.is_encoder_decoder,
            **model_kwargs,
        )
        # init values
        logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
        stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
        logits_warper = logits_warper if logits_warper is not None else model._get_logits_warper(self.config)
        eos_token_id_tensor = torch.tensor([model.config.eos_token_id]).to(input_ids.device)

        # init attention / hidden states / scores tuples
        scores = () if self.config.output_scores else None

        # keep track of which sequences are already finished
        sequences_logscores = torch.zeros(input_ids.shape[0], device=input_ids.device)
        finished = torch.zeros(input_ids.shape[0], device=input_ids.device, dtype=torch.bool)
        
        pad_token_id = pad_token_id if pad_token_id is not None else model.config.pad_token_id

        if not quiet:
            print(f"Running simple ancestral sampling. Batch Size: {input_ids.shape[0]}, "
                  f"logit_processor: {logits_processor}, stopping_criteria: {stopping_criteria}, "
                  f"logits_warper: {logits_warper}")
        while True:
            model_inputs = model.prepare_inputs_for_generation(input_ids, **model_kwargs)
            outputs = model(
                **model_inputs,
                return_dict=True,
            )
            next_token_logits = outputs.logits[:, -1, :]

            next_token_scores = logits_processor(input_ids, next_token_logits)
            next_token_scores = logits_warper(input_ids, next_token_scores)

            if scores is not None:
                scores += (next_token_scores,)
            probs = next_token_scores.softmax(dim=-1)
            next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1) #todo: test and replace this with gumbel sampling
            next_token_logprobs = next_token_scores.log_softmax(dim=-1).gather(dim=-1, index=next_tokens.unsqueeze(-1)).squeeze(-1)
            # next_token_logprobs = torch.where(~finished, next_token_logprobs, torch.zeros_like(next_token_logprobs))
            sequences_logscores += next_token_logprobs * (~finished).float()


            # finished sentences should have their next token be a padding token
            if model.config.eos_token_id is not None:
                next_tokens = next_tokens * ~finished + pad_token_id * finished

            finished = finished | (next_tokens == model.config.eos_token_id)

            # update generated ids, model inputs, and length for next step
            input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
            model_kwargs = model._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=model.config.is_encoder_decoder
            )

            # if eos_token was found in one sentence, set sentence to finished
            if eos_token_id_tensor is not None:
                # unfinished_sequences = unfinished_sequences.mul(
                #     next_tokens.tile(eos_token_id_tensor.shape[0], 1).ne(eos_token_id_tensor.unsqueeze(1)).prod(dim=0)
                # )

                # stop when each sentence is finished
                if (~finished).max() == 0:
                    break

            # stop if we exceed the maximum length
            if stopping_criteria(input_ids, scores):
                break

        return SampleEncoderDecoderOutput(
                sequences=input_ids,
                scores=scores,
                sequences_scores=sequences_logscores,
            )




from decoders import inject_supervitamined_decoders, SmallProbTransformer, SmallProbTransformerConfig, \
    BinaryCodeTransformer, StochasticBeamSearchDecoder
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM


def test_simple_bs_single():
    inputs = ["translate English to German: How old are you?"]
    _test_simple_sbs(inputs)


def test_simple_bs_batch():
    inputs = ["translate English to German: What is your name, my dear Friend? I missed you so much",
        "translate English to German: How old are you?",
        "a b c 1 2 ",
        "summarize: Lorem ipsum dolor "
    ]
    _test_simple_sbs(inputs)


def _test_simple_sbs(inputs):
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    model.eval()
    inject_supervitamined_decoders(model)
    inputs = tokenizer(inputs,
                       return_tensors="pt", padding=True, truncation=True
                       )
    outputs = model.generate(**inputs,
                             generation_strategy=StochasticBeamSearchDecoder(),
                             max_new_tokens=100, num_beams=5, num_return_sequences=5,
                             )

    print(f"generated text: {tokenizer.batch_decode(outputs.sequences, skip_special_tokens=True)}")
    print(f"generated probs: {outputs.sequences_scores}")
    print(f"generated gumbels: {outputs.last_scores}")


def test_binary_transformer():
    import torch
    inputs = torch.tensor([[-1]] * 2)
    model = BinaryCodeTransformer(n=6)
    inject_supervitamined_decoders(model)
    output = model.generate(inputs,
                            generation_strategy=StochasticBeamSearchDecoder(),
                            num_beams=100,
                            num_return_sequences=100,
                            max_new_tokens=100,
                            )
    print(f"generated seqs: {output.sequences}")
    print(f"generated probs: {output.sequences_scores}")
    print(f"generated gumbels: {output.last_scores}")


def test_small_prob_transformer():
    import torch
    torch.manual_seed(43)
    inputs = torch.tensor([[-2]] * 2)
    model = SmallProbTransformer(SmallProbTransformerConfig())
    inject_supervitamined_decoders(model)
    output = model.generate(inputs,
                            generation_strategy=StochasticBeamSearchDecoder(),
                            max_new_tokens=200, num_beams=100, num_return_sequences=100,
                            eval_by_score=True,
                            )
    print(f"generated seqs: {output.sequences}")
    print(f"generated gumbels: {output.last_scores}")
    print(f"generated probs: {output.sequences_scores}")
    routes = output.sequences[:, 1]
    print(f"routes: {routes}")
    print(f"bins: {routes.bincount(minlength=9)}")


if __name__ == '__main__':
    import sys

    def debugger_is_active() -> bool:
        """Return if the debugger is currently active"""
        return hasattr(sys, 'gettrace') and sys.gettrace() is not None

    if debugger_is_active():
        # test_binary_transformer()
        test_small_prob_transformer()
    else:
        from arsenal import testing_framework
        testing_framework(globals())



import torch
from decoders import BinaryCodeTransformer, BeamSearchDecoder, inject_supervitamined_decoders, \
    batched_bin_vec_to_decimal, FakeTransformer, StochasticBeamSearchDecoder


def test_fake_transformer():
    torch.manual_seed(42)
    model = FakeTransformer()
    inject_supervitamined_decoders(model)
    result = model.generate(input_ids=torch.tensor([[0], [0]]),
                            generation_strategy=StochasticBeamSearchDecoder(),
                            num_return_sequences=3,
                            num_beams=3,
                            )
    print(f"generated seqs: {result.sequences}")
    print(f"generated probs: {result.sequences_scores.exp()}")
    assert torch.all(result.sequences == torch.tensor([[0, 1, 2, 3],
                                                          [0, 2, 2, 3],
                                                          [0, 1, 1, 3],
                                                          [0, 2, 2, 3],
                                                          [0, 1, 1, 3],
                                                          [0, 1, 2, 3]]))
    assert torch.allclose(result.sequences_scores.exp(), torch.tensor([0.6750, 0.2250, 0.0750, 0.2250, 0.0750, 0.6750]))


def test_binary_code_transformer_scores():
    torch.manual_seed(0)
    model = BinaryCodeTransformer()
    inject_supervitamined_decoders(model)
    input_ids = torch.tensor([[-1]] * 1000)
    output = model.generate(input_ids, generation_strategy=BeamSearchDecoder())
    numbers = batched_bin_vec_to_decimal(output.sequences[:, 1:])
    analytical_scores = model.p[numbers].log()
    assert torch.allclose(output.sequences_scores.float(), analytical_scores.float())


if __name__ == "__main__":
    from arsenal import testing_framework

    testing_framework(globals())



import numpy as np
import torch
from transformers import PreTrainedModel, PretrainedConfig
from transformers.modeling_outputs import Seq2SeqLMOutput


class BinaryCodeTransformerConfig(PretrainedConfig):
    """Configuration class with appropriate defaults for BinaryCodeTransformer."""
    model_type = "BinaryCodeTransformer"

    def __init__(self, max_new_tokens=100, vocab_size=3, **kwargs):
        super().__init__(pad_token_id=-5, eos_token_id=2, bos_token_id=-1, **kwargs)
        self.vocab_size = vocab_size
        self.max_new_tokens = max_new_tokens
        self.do_sample = True
        self.num_beams = 1


# Helper functions

@torch.jit.script
def batched_bin_vec_to_decimal(vec):
    """Convert rows with binary digits to decimal natural numbers.
    Example: tensor([[0, 1, 0, 1, 1, 0]) -> tensor([22])"""
    if torch.all(vec[..., -1] == 2):
        vec = vec[..., :-1]
    return torch.matmul(vec.long(), (2 ** torch.arange(vec.size(-1) - 1, -1, -1))).long()


def decimal_to_bin_vec(num, n):
    """Convert decimal natural number to binary vector."""
    tensor_without_eos = torch.tensor(list(map(int, bin(num)[2:].zfill(n))))
    return torch.cat((tensor_without_eos, torch.tensor([2])))


def get_binary_completions(seq, max_len):
    """Get all binary completions of a sequence up to a given length.
    Example: seq=[0, 1], max_len=4 -> [[0, 1, 0, 0], [0, 1, 0, 1], [0, 1, 1, 0], [0, 1, 1, 1]]"""
    seq_len = seq.shape[-1]
    empty_bits = max_len - seq_len
    seqs = torch.zeros(2 ** empty_bits, max_len)
    seqs[:, :seq_len] = seq.repeat(1, 2 ** empty_bits, 1)
    seqs[:, seq_len:] = torch.tensor([list(map(int, bin(i)[2:].zfill(empty_bits))) for i in range(2 ** empty_bits)])
    return seqs


def hash_seq(seq):
    """Hash a sequence of binary digits."""
    return hash(tuple(seq.reshape(-1).tolist()))
    # return batched_bin_vec_to_decimal(seq).item()*1000 + seq.shape[-1]


class BinaryCodeTransformer(PreTrainedModel):
    """A model stub that samples binary sequences
    of a fixed length n with a predetermined discrete
    distribution over all 2^n possible sequences.

    The model maps each sequence to its decimal
    representation as an integer from 0 to 2^n-1"""
    config_class = BinaryCodeTransformerConfig

    def __init__(self, config=None, n=6, prob_vector=None):
        """Initialize the model.

        :param config: BinaryCodeTransformerConfig
        :param n: length of binary sequences (there will be 2^n possible sequences)
        :param prob_vector: a vector of probabilities over all possible sequences,
                            indexed by their decimal representation
        """
        config = BinaryCodeTransformerConfig() if config is None else config
        super().__init__(config)
        self.fake_param = torch.nn.Parameter(torch.tensor(1.0))  # need at least one parameter to be a valid model
        self.n = n
        self.num_seqs = 2 ** self.n
        if prob_vector is not None:
            assert len(prob_vector) == self.num_seqs
            assert torch.allclose(torch.tensor(prob_vector).sum(), torch.tensor(1.0))
            self.p = torch.tensor(prob_vector, dtype=torch.float32)
        else:
            self.p = torch.tensor(np.random.dirichlet(np.ones(self.num_seqs)), dtype=torch.float32)
        self.prob_mem = {}
        self.precompute_probs()

    def precompute_probs(self):
        """Precompute probabilities of all possible sequences"""
        seq = torch.tensor([[-1]])
        seqs = [s for i in range(2, self.n + 2) for s in get_binary_completions(seq, i)]
        probs = [self.seq_prob(seq[1:]) for seq in seqs]
        for seq, prob in zip(seqs, probs):
            self.prob_mem[hash_seq(seq[1:])] = prob

    def seq_prob(self, seq):
        """Compute the probability of a binary sequence.
        When it is incomplete, sum over all possible completions."""
        assert len(seq.shape) == 1
        seq_len = seq.shape[-1]
        if seq_len == self.n:
            value = self.p[batched_bin_vec_to_decimal(seq)]
        else:
            seqs = get_binary_completions(seq, self.n)
            values = batched_bin_vec_to_decimal(seqs)
            value = self.p[values].sum()
        return value

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            encoder_outputs=None,
            past_key_values=None,
            return_dict=True,
            output_attentions=None,
            output_hidden_states=None,
            **kwargs,
    ):
        assert past_key_values is None
        batch_size, seq_len = input_ids.shape

        # append an extra middle dim for the vocab size
        def next_seqs(seqs):
            expanded_t = seqs.unsqueeze(1).expand(-1, 2, -1)
            new_dim = torch.zeros(batch_size, 2, 1)
            new_dim[:, 0, :] = 0
            new_dim[:, 1, :] = 1
            return torch.cat((expanded_t, new_dim), dim=2)

        if input_ids.shape[1] == self.n + 1:
            output = torch.zeros(batch_size, 1, 3)
            output[:, 0, 2] = 1.0
        else:
            continuations = next_seqs(input_ids).view(batch_size * 2, seq_len + 1)
            continuations = continuations[:, 1:].unbind()
            continuations_probs = torch.tensor([self.prob_mem[hash_seq(seq)] for seq in continuations],
                                               dtype=torch.float32)
            # output[:, seq_len - 1, :] = continuations_probs.view(batch_size, self.config.vocab_size)
            output = continuations_probs.view(batch_size, 2).unsqueeze(1)
            output = torch.cat([output, torch.zeros(batch_size, 1, 1)], dim=-1)
        # Placeholder for past_key_values, attentions, hidden_states
        past_key_values = () if past_key_values is None else past_key_values
        attentions = () if output_attentions else None
        hidden_states = () if output_hidden_states else None

        output = output.log()
        # replace -inf with -1e9
        # output[output == float('-inf')] = -1e9

        # Create Seq2SeqLMOutput object
        if not return_dict:
            att, hidd = (
                attentions if attentions is not None else (),
                hidden_states if hidden_states is not None else (),
            )
            return (output,) + past_key_values + hidd + att
        else:
            return Seq2SeqLMOutput(
                loss=None,
                logits=output,
                past_key_values=past_key_values,
                decoder_hidden_states=hidden_states,
                decoder_attentions=attentions,
                encoder_last_hidden_state=None,
                encoder_hidden_states=None,
                encoder_attentions=None,
            )

    def prepare_inputs_for_generation(self, input_ids, **kwargs):
        return {"input_ids": input_ids}

    def _reorder_cache(self, past, beam_idx):
        return past


def demo():
    config = BinaryCodeTransformerConfig(max_new_tokens=6)
    model = BinaryCodeTransformer(config)

    input_ids = torch.tensor([[-1]] * 10)
    output = model.generate(
        input_ids,
        return_dict_in_generate=True,
        output_scores=True,
    )
    print(f"generated tokens: {output.sequences}")
    print(f"probs of seq0 at step 1: {torch.exp(output.scores[0][0, :])}")
    print(f"probs of seq0 at step 2: {torch.exp(output.scores[1][0, :])}")
    print(f"probs of seq0 at step 3: {torch.exp(output.scores[2][0, :])}")
    print(f"probs of seq0 at step 4: {torch.exp(output.scores[3][0, :])}")
    print(f"probs of seq0 at step 5: {torch.exp(output.scores[4][0, :])}")
    print(f"probs of seq0 at step 6: {torch.exp(output.scores[5][0, :])}")
    print(f"probs predetermined seq0: {model.p[batched_bin_vec_to_decimal(output.sequences[0])]}")


def test():
    torch.manual_seed(0)
    np.random.seed(0)
    config = BinaryCodeTransformerConfig(max_new_tokens=6)
    model = BinaryCodeTransformer(config)

    input_ids = torch.tensor([[-1]] * 10)
    output = model.generate(input_ids, return_dict_in_generate=True, output_scores=True)
    torch.set_printoptions(precision=8)
    assert torch.allclose(
        model.p,
        torch.tensor(
            [0.01277674, 0.02016235, 0.01482116, 0.01263750, 0.00884638, 0.01666631, 0.00923922, 0.03569583,
                0.05321665, 0.00776361, 0.02518661, 0.01208322, 0.01347602, 0.04171163, 0.00118292, 0.00146347,
                0.00032791, 0.02869582, 0.02417343, 0.03275462, 0.06173009, 0.02577006, 0.00993612, 0.02434604,
                0.00202075, 0.01639778, 0.00248399, 0.04646620, 0.01184487, 0.00859782, 0.00493299, 0.02389202,
                0.00977804, 0.01349049, 0.00030452, 0.01543374, 0.01520281, 0.01540431, 0.04620123, 0.01838373,
                0.00715223, 0.00922338, 0.01920197, 0.00099718, 0.01764163, 0.01782921, 0.00379200, 0.00221587,
                0.00608375, 0.00725792, 0.01355620, 0.00926820, 0.07151123, 0.00172794, 0.00376141, 0.00282407,
                0.01699675, 0.00468898, 0.01008079, 0.00449949, 0.00277934, 0.00187757, 0.01714653, 0.00238738,
            ], dtype=torch.float32),
    )
    print(f"scores: {output.scores[5]}")
    assert torch.allclose(
        output.scores[5],
        torch.tensor(
            [
                [-5.58296156, -5.86957741],  # each column is a different sample
                [-2.78498363, -3.65854216],
                [-2.63790083, -6.36082315],
                [-4.61157894, -3.71538639],
                [-4.18627453, -4.17310810],
                [-4.03749371, -4.02691698],
                [-4.43586063, -4.75624657],
                [-3.68144274, -4.41593790],
                [-4.30684376, -3.17697525],
                [-8.02278233, -3.55100393],
            ],
            dtype=torch.float32,
        ),
    )
    print("test passed")


def benchmark():
    import timeit
    from .fake_transformer import FakeTransformerConfig
    from .fake_transformer import FakeTransformer

    models = {
        "BinaryCodeTransformer": BinaryCodeTransformer(BinaryCodeTransformerConfig(max_new_tokens=6)),
        "FakeTransformer": FakeTransformer(FakeTransformerConfig()),
    }
    inputs = {
        "BinaryCodeTransformer": torch.tensor([[-1]] * 1000),
        "FakeTransformer": torch.tensor([[0]] * 1000),
    }

    def run(model, inputs):
        model.generate(
            inputs,
            return_dict_in_generate=True,
            output_scores=True,
        )

    for name, model in models.items():
        print(f"{name}: {timeit.timeit(lambda: run(model, inputs[name]), number=10)}")


if __name__ == "__main__":
    import typer

    app = typer.Typer()
    app.command()(demo)
    app.command()(benchmark)
    app.command()(test)
    app()



from transformers import AutoTokenizer, AutoModelForSeq2SeqLM

from decoders import inject_supervitamined_decoders
from decoders.simple.ancestral import SamplingDecoder
from decoders.toolbox import compute_logprobs_from_scores, compute_true_logprobs


def test_t5_small_ancestral_noscores():
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    inject_supervitamined_decoders(model)
    inputs = tokenizer([  # "translate English to German: What is your name, my dear Friend? I missed you so much",
        "translate English to German: How old are you?",
    ],
        return_tensors="pt", padding=True, truncation=True
    )
    outputs = model.generate(**inputs, generation_strategy=SamplingDecoder(), max_new_tokens=100,
                             top_k=0, num_return_sequences=100, do_sample=True, output_scores=False)
    gen_logp = outputs.sequences_scores
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences, skip_special_tokens=True)}")
    print(f"generated tokens: {outputs.sequences}")
    print(f"generated logp: {gen_logp}")
    # assert outputs.scores is None
    # _, recompute_logp = compute_true_logprobs(model, outputs.sequences, encoder_input=inputs)
    # print(f"true logp: {recompute_logp.sum(dim=1)}")


def test_t5_unicamp_ancestral_noscores():
    tokenizer = AutoTokenizer.from_pretrained("unicamp-dl/translation-en-pt-t5")
    model = AutoModelForSeq2SeqLM.from_pretrained("unicamp-dl/translation-en-pt-t5")
    inject_supervitamined_decoders(model)
    inputs = tokenizer([  # "translate English to German: What is your name, my dear Friend? I missed you so much",
        "translate English to German: How old are you?",
    ], return_tensors="pt", padding=True, truncation=True)
    outputs = model.generate(**inputs, generation_strategy=SamplingDecoder(),
                             max_new_tokens=100, top_k=0, num_return_sequences=100,
                             do_sample=True, output_scores=False)
    # num_beams=1, do_sample=1, top_k=0,
    # return_dict_in_generate=True, output_scores=True)
    gen_logp = outputs.sequences_scores
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences)}")
    print(f"generated tokens: {outputs.sequences}")
    print(f"generated logp: {gen_logp}")
    _, recompute_logp = compute_true_logprobs(model, outputs.sequences, encoder_input=inputs)
    print(f"true logp: {recompute_logp.sum(dim=1)}")


def test_ancestral():
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    inject_supervitamined_decoders(model)
    inputs = tokenizer(["translate English to German: What is your name, my dear Friend? I missed you so much",
                           "translate English to German: How old are you?",
                       ], return_tensors="pt", padding=True, truncation=True)
    outputs = model.generate(**inputs, generation_strategy=SamplingDecoder(),
                             max_new_tokens=100, top_k=0, do_sample=True, output_scores=True)
    gen_logp = outputs.sequences_scores
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences)}")
    print(f"generated tokens: {outputs.sequences}")
    print(f"generated logp: {gen_logp}")
    logp_computed = compute_logprobs_from_scores(model, outputs)
    print(f"compute_transition logp: {logp_computed.sum(dim=1)}")
    _, recompute_logp = compute_true_logprobs(model, outputs.sequences, encoder_input=inputs)
    print(f"true logp: {recompute_logp.sum(dim=1)}")


if __name__ == '__main__':
    from arsenal import testing_framework

    testing_framework(globals())



import torch
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, NoBadWordsLogitsProcessor
from decoders import inject_supervitamined_decoders, BeamSearchDecoder, LogitsProcessorList
import time


def test_simple_bs_single():
    inputs = ["translate English to German: How old are you?"]
    _test_simple_beam_search(inputs)


def test_simple_bs_batch():
    inputs = ["translate English to German: What is your name, my dear Friend? I missed you so much",
        "translate English to German: How old are you?",
        "a b c 1 2 ",
        "summarize: Lorem ipsum dolor "
    ]
    _test_simple_beam_search(inputs)


def _test_simple_beam_search(inputs):
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    inject_supervitamined_decoders(model)
    inputs = tokenizer(inputs, return_tensors="pt", padding=True, truncation=True)
    t0 = time.time()
    outputs = model.generate(**inputs,
                             generation_strategy=BeamSearchDecoder(),
                             max_new_tokens=100, num_beams=5,
                             keep_k_always_alive=True,
                             )
    t1 = time.time()
    gold_seqs = model.generate(**inputs,
                               num_beams=5, do_sample=False, max_new_tokens=100, length_penalty=0.0,
                               early_stopping=True, return_dict_in_generate=True, output_scores=True)
    t2 = time.time()
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences, skip_special_tokens=True)}")
    print(f"gold text: {tokenizer.batch_decode(gold_seqs.sequences, skip_special_tokens=True)}")
    print(f"generated probs: {outputs.sequences_scores}")
    print(f"gold probs: {gold_seqs.sequences_scores}")
    print(f"simple took {t1 - t0:.2f}s, gold took {t2 - t1:.2f}s")
    assert torch.all(
        outputs.sequences == gold_seqs.sequences), f"output: {outputs.sequences}, gold: {gold_seqs.sequences}"


def test_logit_processor():
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelForSeq2SeqLM.from_pretrained("t5-small")
    inject_supervitamined_decoders(model)
    inputs = tokenizer(["translate English to German: How old are you?", ],
                       return_tensors="pt", padding=True, truncation=True
                       )
    outputs = model.generate(**inputs,
                             generation_strategy=BeamSearchDecoder(),
                             max_new_tokens=100, num_beams=5,
                             logits_processor=LogitsProcessorList(
                                 [NoBadWordsLogitsProcessor(bad_words_ids=[[292]], eos_token_id=1)]),
                             keep_k_always_alive=False)
    print(f"generated logp: {outputs.sequences_scores}")
    print(f"generated tokens: {outputs.sequences}")
    print(f"generated text: {tokenizer.batch_decode(outputs.sequences, skip_special_tokens=True)}")
    assert torch.all(outputs.sequences == torch.tensor([[0, 2739, 4445, 15840, 146, 58, 1]]))


def test_en_pt_model():
    tokenizer = AutoTokenizer.from_pretrained("unicamp-dl/translation-en-pt-t5")
    model = AutoModelForSeq2SeqLM.from_pretrained("unicamp-dl/translation-en-pt-t5")
    inject_supervitamined_decoders(model)

    source_sent = "This question should also provide information regarding the preconditions for the origins of life."
    source_tok = tokenizer(source_sent, return_tensors="pt")
    from decoders.simple.beam_search import BeamSearchDecoder
    result = model.generate(
        source_tok["input_ids"],
        generation_strategy=BeamSearchDecoder(),
        num_beams=100,
        num_return_sequences=100,
        max_length=100,
    )
    print(tokenizer.batch_decode(result.sequences, skip_special_tokens=True))
    print(result.sequences)
    print(result.sequences_scores)

def test_mixtral():
    from transformers import MixtralForCausalLM
    
    dummy_input = torch.LongTensor([[0, 0, 0, 0, 0, 0, 1, 2, 3], 
                                    [1, 1, 2, 3, 4, 5, 6, 7, 8],
                                    [0, 0, 0, 0, 0, 0, 1, 2, 3],])
                                    
    attention_mask = dummy_input.ne(0).to(torch.long)

    model = MixtralForCausalLM.from_pretrained("hf-internal-testing/Mixtral-tiny", torch_dtype=torch.bfloat16, low_cpu_mem_usage=True)
    
    inject_supervitamined_decoders(model)

    from decoders.simple.beam_search import BeamSearchDecoder
    result = model.generate(
        dummy_input,
        attention_mask=attention_mask,
        generation_strategy=BeamSearchDecoder(),
        # num_beams=100,
        # num_return_sequences=100,
        max_length=100,
    )
    print(result.sequences)
    print(result.sequences_scores)

if __name__ == '__main__':
    import sys

    def debugger_is_active() -> bool:
        """Return if the debugger is currently active"""
        return hasattr(sys, 'gettrace') and sys.gettrace() is not None

    if debugger_is_active():
        # test_binary_transformer()
        test_mixtral()
    else:
        from arsenal import testing_framework
        testing_framework(globals())



import torch
import os
import torch.nn.functional as F


def gumbel_like(*args, **kwargs):
    return _gumbel(torch.rand_like(*args, **kwargs))


def gumbel(*args, **kwargs):
    return _gumbel(torch.rand(*args, **kwargs))


def _gumbel(u):
    return -torch.log(-torch.log(u))


def gumbel_log_survival(x):
    """Computes log P(g > x) = log(1 - P(g < x)) = log(1 - exp(-exp(-x))) for a standard Gumbel"""
    y = torch.exp(-x)
    return torch.where(
        x >= 10,  # means that y < 1e-4 so O(y^6) <= 1e-24 so we can use series expansion
        -x - y / 2 + y ** 2 / 24 - y ** 4 / 2880,  # + O(y^6), https://www.wolframalpha.com/input/?i=log(1+-+exp(-y))
        torch.log(-torch.expm1(-torch.exp(-x)))  # Hope for the best
    )

GUMBEL_SENSITIVITY = float(os.environ.get('GUMBEL_SENSITIVITY', 1e-3))
if os.environ.get('GUMBEL_SENSITIVITY'):
    print(f"Gumbel sensitivity set to {GUMBEL_SENSITIVITY :.2e}")



def gumbel_with_maximum(phi, T, dim=-1):
    """
    Samples a set of gumbels which are conditioned on having a maximum along a dimension
    phi.max(dim)[0] should be broadcastable with the desired maximum T
    """

    # if ((-1e9 < phi) & (phi< -100)).any():
    #     print("habemus problemus")

    NEG_INF_THRESHOLD = -float('inf')

    # phi = phi.clamp(min=-1e10)
    inf_rows_mask = (phi == NEG_INF_THRESHOLD).all(dim=1)
    if inf_rows_mask.any():
        old_phi = phi.clone()
        phi = phi[~inf_rows_mask]
        old_T = T.clone()
        T = T[~inf_rows_mask]

    # Gumbel with location phi
    g_phi = phi + gumbel_like(phi)
    Z, argmax = g_phi.max(dim)
    g = _shift_gumbel_maximum(g_phi, T, dim, Z=Z)
    CHECK_VALIDITY = True
    if CHECK_VALIDITY:
        g_inv = _shift_gumbel_maximum(g, Z, dim)
        # Create a boolean mask where the condition fails
        mask = ~(((g_phi - g_inv) < GUMBEL_SENSITIVITY) | (g_phi == g_inv))
        # | torch.isinf(g_phi).all(dim).repeat_interleave(g_phi.shape[-1], dim).view(g_phi.shape))

        # Get indices where the mask is True (i.e., the assertion fails)
        fail_indices = torch.nonzero(mask, as_tuple=True)

        # Print the failed indices and corresponding values
        for ind in zip(*fail_indices):
            print(f"Index: {ind}, phi: {phi[ind]}, g_phi: {g_phi[ind]}, g_inv: {g_inv[ind]}")

        assert (((g_phi - g_inv) < GUMBEL_SENSITIVITY) | (g_phi == g_inv)).all()
        # | torch.isinf(g_phi).all(dim).repeat_interleave(g_phi.shape[-1], dim).view(g_phi.shape)).all()
        # if a row in g_phi is all -inf, set the corresponding row in g to -inf
        # g[torch.isinf(g_phi).all(dim)] = float('-inf')
    if inf_rows_mask.any():
        g_new = torch.full_like(old_phi, NEG_INF_THRESHOLD, device=old_phi.device)
        g_new[~inf_rows_mask] = g
        g = g_new
        argmax_new = torch.full_like(old_T, -1,dtype=torch.long, device=old_T.device)
        argmax_new[~inf_rows_mask] = argmax
        argmax = argmax_new
    return g, argmax


def _shift_gumbel_maximum(g_phi, T, dim=-1, Z=None):
    if Z is None:
        Z, _ = g_phi.max(dim)
    u = T.unsqueeze(dim) - g_phi + torch.log1p(-torch.exp(g_phi - Z.unsqueeze(dim)))
    return T.unsqueeze(dim) - F.relu(u) - torch.log1p(torch.exp(-u.abs()))



import torch
from decoders import inject_supervitamined_decoders, SmallProbTransformer, SmallProbTransformerConfig, SamplingDecoder

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

config = SmallProbTransformerConfig()
model = SmallProbTransformer(config).to(device)
inject_supervitamined_decoders(model)


def test_ancestral_sample():
    torch.manual_seed(43)
    input_ids = torch.tensor([[-1]]).to(device)
    output = model.generate(input_ids,
                            generation_strategy=SamplingDecoder(),
                            max_new_tokens=200, top_k=0, do_sample=True,
                            num_return_sequences=20, )
    routes = output.sequences[:, 1]
    bins = routes.bincount(minlength=config.real_vocab_size)
    print(f"generated tokens: {output.sequences}")
    print(f"routes: {routes}")
    print(f"bins: {bins}")
    print(f"generated probs: {output.sequences_scores}")


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())